@@ -1,6 +1,25 @@
 /*
- * Read log messages from the kernel - send them to syslogd
- *
+    klogd.c - main program for Linux kernel log daemon.
+    Copyright (c) 1995  Dr. G.W. Wettstein <greg@wind.rmcc.com>
+
+    This file is part of the sysklogd package, a kernel and system log daemon.
+
+    This program is free software; you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation; either version 2 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program; if not, write to the Free Software
+    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+*/
+
+/*
  * Steve Lord (lord@cray.com) 7th Nov 92
  *
  * Modified to check for kernel info by Dr. G.W. Wettstein 02/17/93.
@@ -65,19 +84,78 @@
  *	Added pid saving when the daemon detaches into the background.  Thank
  *	you to Juha Virtanen (jiivee@hut.fi) for providing this patch.
  *
+ * Mon Feb  6 07:31:29 CST 1995:  Dr. Wettstein
+ *	Significant re-organization of the signal handling code.  The
+ *	signal handlers now only set variables.  Not earth shaking by any
+ *	means but aesthetically pleasing to the code purists in the group.
+ *
+ *	Patch to make things more compliant with the file system standards.
+ *	Thanks to Chris Metcalf for prompting this helpful change.
+ *
+ *	The routines responsible for reading the kernel log sources now
+ *	initialize the buffers before reading.  I think that this will
+ *	solve problems with non-terminated kernel messages producing
+ *	output of the form:  new old old old
+ *
+ *	This may also help influence the occassional reports of klogd
+ *	failing under significant load.  I think that the jury may still
+ *	be out on this one though.  My thanks to Joerg Ahrens for initially
+ *	tipping me off to the source of this problem.  Also thanks to
+ *	Michael O'Reilly for tipping me off to the best fix for this problem.
+ *	And last but not least Mark Lord for prompting me to try this as
+ *	a means of attacking the stability problem.
+ *
+ *	Specifying a - as the arguement to the -f switch will cause output
+ *	to be directed to stdout rather than a filename of -.  Thanks to
+ *	Randy Appleton for a patch which prompted me to do this.
+ *
+ * Wed Feb 22 15:37:37 CST 1995:  Dr. Wettstein
+ *	Added version information to logging startup messages.
+ *
+ * Wed Jul 26 18:57:23 MET DST 1995: Martin Schulze
+ *	Added an commandline argument "-n" to avoid forking. This obsoletes
+ *	the compiler define NO_FORK. It's more useful to have this as an
+ *	argument as there are many binary versions and one doesn't need to
+ *	recompile the daemon.
+ *
+ * Thu Aug 10 19:01:08 MET DST 1995: Martin Schulze
+ *	Added my pidfile.[ch] to it to perform a better handling with pidfiles.
+ *	Now both, syslogd and klogd, can only be started once. They check the
+ *	pidfile.
+ *
+ * Fri Nov 17 15:05:43 CST 1995:  Dr. Wettstein
+ *	Added support for kernel address translation.  This required moving
+ *	some definitions and includes to the new klogd.h file.  Some small
+ *	code cleanups and modifications.
+ *
+ * Mon Nov 20 10:03:39 MET 1995
+ *	Added -v option to print the version and exit.
+ *
+ * Thu Jan 18 11:19:46 CST 1996:  Dr. Wettstein
+ *	Added suggested patches from beta-testers.  These address two
+ *	two problems.  The first is segmentation faults which occur with
+ *	the ELF libraries.  This was caused by passing a null pointer to
+ *	the strcmp function.
+ *
+ *	Added a second patch to remove the pidfile as part of the
+ *	termination cleanup sequence.  This minimizes the potential for
+ *	conflicting pidfiles causing immediate termination at boot time.
+ *	
  */
 
 
-#include <stdio.h>
+/* Includes. */
 #include <unistd.h>
 #include <signal.h>
-#include <string.h>
-#include <syslog.h>
 #include <errno.h>
 #include <sys/fcntl.h>
 #include <sys/stat.h>
 #include <linux/time.h>
 #include <stdarg.h>
+#include <paths.h>
+#include "klogd.h"
+#include "pidfile.h"
+#include "version.h"
 
 #define __LIBRARY__
 #include <linux/unistd.h>
@@ -86,17 +164,22 @@ _syscall3(int,sys_syslog,int, type, char *, buf, int, len);
 
 #define LOG_BUFFER_SIZE 4096
 #define LOG_LINE_LENGTH 1024
-#define PROC_KMSG "/proc/kmsg"
 
-static char	*PidFile = "/var/run/klogd.pid";
+#if defined(FSSTND)
+static char	*PidFile = _PATH_VARRUN "klogd.pid";
+#else
+static char	*PidFile = "/etc/klogd.pid";
+#endif
 
 static int	kmsg,
+		change_state = 0,
+		terminate = 0,
 		caught_TSTP = 0,
-		console_log_level = 1;
+		console_log_level = 6;
 
 static int	use_syscall = 0,
 		one_shot = 0,
-		debugging = 0;
+		NoFork = 0;	/* don't fork - don't run in daemon mode */
 
 static char log_buffer[LOG_BUFFER_SIZE];
 
@@ -104,21 +187,25 @@ static FILE *output_file = (FILE *) 0;
 
 static enum LOGSRC {none, proc, kernel} logsrc;
 
-static enum LOGSRC GetKernelLogSrc(void);
+int debugging = 0;
 
 
 /* Function prototypes. */
-void CloseLogSrc(void);
+extern int sys_syslog(int type, char *buf, int len);
+static void CloseLogSrc(void);
+extern void restart(int sig);
+extern void stop_logging(int sig);
+extern void stop_daemon(int sig);
+static void Terminate(void);
+static void ChangeLogging(void);
 static enum LOGSRC GetKernelLogSrc(void);
-static void Syslog(int, char *, ...);
 static void LogLine(char *ptr, int len);
 static void LogKernelLine(void);
 static void LogProcLine(void);
-int main(int argc, char *argv[]);
-int sys_syslog(int type, char *buf, int len);
+extern int main(int argc, char *argv[]);
 
 
-void CloseLogSrc()
+static void CloseLogSrc()
 
 {
 	/* Turn on logging of messages to console. */
@@ -145,45 +232,92 @@ void CloseLogSrc()
 }
 
 
-void restart()
+void restart(sig)
+	
+	int sig;
 
 {
 	signal(SIGCONT, restart);
-	if ( caught_TSTP != 1 )
-	{
-		Syslog(LOG_INFO, "Kernel logging re-started after SIGSTOP.");
-		return;
-	}
-	
-	CloseLogSrc();
-	logsrc = GetKernelLogSrc();
+	change_state = 1;
 	caught_TSTP = 0;
 	return;
 }
 
 
-void stop_logging()
+void stop_logging(sig)
+
+	int sig;
+	
 {
 	signal(SIGTSTP, stop_logging);
-	CloseLogSrc();
+	change_state = 1;
 	caught_TSTP = 1;
-	logsrc = none;
 	return;
 }
 
 
-void stop_daemon(int sig)
+void stop_daemon(sig)
+
+	int sig;
 
 {
-  	CloseLogSrc();
- 	Syslog(LOG_INFO, "Kernel log daemon terminating.");
- 	sleep(1);
+	change_state = 1;
+	terminate = 1;
+	return;
+}
+
+
+static void Terminate()
+
+{
+	CloseLogSrc();
+	Syslog(LOG_INFO, "Kernel log daemon terminating.");
+	sleep(1);
 	if ( output_file != (FILE *) 0 )
 		fclose(output_file);
 	closelog();
+	(void) remove_pid(PidFile);
 	exit(1);
 }
 
+	
+static void ChangeLogging(void)
+
+{
+	/* Terminate kernel logging. */
+	if ( terminate == 1 )
+		Terminate();
+
+	/* Stop kernel logging. */
+	if ( caught_TSTP == 1 )
+	{
+		CloseLogSrc();
+		logsrc = none;
+		change_state = 0;
+		return;
+	}
+		
+	/*
+	 * The rest of this function is responsible for restarting
+	 * kernel logging after it was stopped.
+	 *
+	 * In the following section we make a decision based on the
+	 * kernel log state as to what is causing us to restart.  Somewhat
+	 * groady but it keeps us from creating another static variable.
+	 */
+	if ( logsrc != none )
+	{
+		Syslog(LOG_INFO, "Kernel logging re-started after SIGSTOP.");
+		change_state = 0;
+		return;
+	}
+
+	/* Restart logging. */
+	logsrc = GetKernelLogSrc();
+	change_state = 0;
+	return;
+}
+
 
 static enum LOGSRC GetKernelLogSrc(void)
 
@@ -211,27 +345,30 @@ static enum LOGSRC GetKernelLogSrc(void)
 	 * First do a stat to determine whether or not the proc based
 	 * file system is available to get kernel messages from.
 	 */
-	if ( use_syscall || ((stat(PROC_KMSG, &sb) < 0) && (errno == ENOENT)) )
+	if ( use_syscall ||
+	    ((stat(_PATH_KLOG, &sb) < 0) && (errno == ENOENT)) )
 	{
 	  	/* Initialize kernel logging. */
 	  	sys_syslog(1, NULL, 0);
-		Syslog(LOG_INFO, "Kernel logging (sys_syslog) started.");
+		Syslog(LOG_INFO, "klogd %s-%s, log source = sys_syslog "
+		       "started.", VERSION, PATCHLEVEL);
 		return(kernel);
 	}
 	
-	if ( (kmsg = open(PROC_KMSG, O_RDONLY)) < 0 )
+	if ( (kmsg = open(_PATH_KLOG, O_RDONLY)) < 0 )
 	{
 		fputs("klogd: Cannot open proc file system.", stderr);
 		sys_syslog(7, NULL, 0);
 		exit(1);
 	}
 
-	Syslog(LOG_INFO, "Kernel logging (proc) started.");
+	Syslog(LOG_INFO, "klogd %s-%s, log source = %s started.", \
+	       VERSION, PATCHLEVEL, _PATH_KLOG);
 	return(proc);
 }
 
 
-static void Syslog(int priority, char *fmt, ...)
+extern void Syslog(int priority, char *fmt, ...)
 
 {
 	va_list ap;
@@ -251,6 +388,7 @@ static void Syslog(int priority, char *fmt, ...)
 		va_end(ap);
 		fputc('\n', output_file);
 		fflush(output_file);
+		fsync(fileno(output_file));
 		return;
 	}
 	
@@ -301,7 +439,9 @@ static void LogLine(char *ptr, int len)
 	auto int idx = 0;
 	static int index = 0;
 	auto char *nl;
-	static char line[LOG_LINE_LENGTH];
+	static char line[LOG_LINE_LENGTH],
+		    eline[LOG_LINE_LENGTH];
+
 
 	if ( debugging && (len != 0) )
 	{
@@ -336,11 +476,16 @@ static void LogLine(char *ptr, int len)
 			   kernel messages have multiple terminators, eg.
 			   \n\r) */
 			if ( (*line != '\n') && (*line != '\r') )
-				Syslog(LOG_INFO, line);
-		      index = 0;
-		      memset(line, '\0', sizeof(line));
+			{
+				memset(eline, '\0', sizeof(eline));
+				ExpandKadds(line, eline);
+				Syslog(LOG_INFO, eline);
+			}
+			index = 0;
+			memset(line, '\0', sizeof(line));
 		 }
-		 else{
+		 else
+		 {
 			 if ( debugging )
 			 {
 				 fputs("No terminator - leftover:\n", stderr);
@@ -363,9 +508,16 @@ static void LogLine(char *ptr, int len)
 static void LogKernelLine(void)
 
 {
-	auto int len;
+	auto int rdcnt;
 
-	if ( (len = sys_syslog(2, log_buffer, sizeof(log_buffer))) < 0 )
+	/*
+	 * Zero-fill the log buffer.  This should cure a multitude of
+	 * problems with klogd logging the tail end of the message buffer
+	 * which will contain old messages.  Then read the kernel log
+	 * messages into this fresh buffer.
+	 */
+	memset(log_buffer, '\0', sizeof(log_buffer));
+	if ( (rdcnt = sys_syslog(2, log_buffer, sizeof(log_buffer))) < 0 )
 	{
 		if ( errno == EINTR )
 			return;
@@ -373,7 +525,7 @@ static void LogKernelLine(void)
 			errno, strerror(errno));
 	}
 	
-	LogLine(log_buffer, len);
+	LogLine(log_buffer, rdcnt);
 	return;
 }
 
@@ -383,6 +535,13 @@ static void LogProcLine(void)
 {
 	auto int rdcnt;
 
+	/*
+	 * Zero-fill the log buffer.  This should cure a multitude of
+	 * problems with klogd logging the tail end of the message buffer
+	 * which will contain old messages.  Then read the kernel messages
+	 * from the message pseudo-file into this fresh buffer.
+	 */
+	memset(log_buffer, '\0', sizeof(log_buffer));
 	if ( (rdcnt = read(kmsg, log_buffer, sizeof(log_buffer))) < 0 )
 	{
 		if ( errno == EINTR )
@@ -403,15 +562,14 @@ int main(argc, argv)
 	char *argv[];
 
 {
-	auto int ch;
+	auto int ch, use_output = 0;
 
-	auto char *log_level = (char *) 0,
-		  *output = (char *) 0;
-
-	auto FILE *fp;
+	auto char	*symfile = (char *) 0,
+			*log_level = (char *) 0,
+			*output = (char *) 0;
 
 	/* Parse the command-line. */
-	while ((ch = getopt(argc, argv, "c:df:os")) != EOF)
+	while ((ch = getopt(argc, argv, "c:df:k:nosv")) != EOF)
 		switch((char)ch)
 		{
 		    case 'c':		/* Set console message level. */
@@ -422,6 +580,13 @@ int main(argc, argv)
 			break;
 		    case 'f':		/* Define an output file. */
 			output = optarg;
+			use_output++;
+			break;
+		    case 'k':		/* Kernel symbol file. */
+			symfile = optarg;
+			break;
+		    case 'n':		/* don't fork */
+			NoFork++;
 			break;
 		    case 'o':		/* One-shot mode. */
 			one_shot = 1;
@@ -429,6 +594,9 @@ int main(argc, argv)
 		    case 's':		/* Use syscall interface. */
 			use_syscall = 1;
 			break;
+		    case 'v':
+			printf("klogd %s-%s\n", VERSION, PATCHLEVEL);
+			exit (1);
 		}
 
 
@@ -452,25 +620,55 @@ int main(argc, argv)
 	 * The child closes all its open file descriptors, and issues a
 	 * call to setsid to establish itself as an independent session
 	 * immune from control signals.
+	 *
+	 * fork() is only called if it should run in daemon mode, fork is
+	 * not disabled with the command line argument and there's no
+	 * such process running.
 	 */
-#if !defined(NO_FORK)
-	if ( !one_shot )
+	if ( (!one_shot) && (!NoFork) )
 	{
-		if ( fork() == 0 )
+		if (!check_pid(PidFile))
 		{
-			auto int fl;
+			if ( fork() == 0 )
+			{
+				auto int fl;
+				int num_fds = getdtablesize();
 		
-			/* This is the child closing its file descriptors. */
-			for (fl= 0; fl <= FD_SETSIZE; ++fl)
-				close(fl);
-			setsid();
+				/* This is the child closing its file descriptors. */
+				for (fl= 0; fl <= num_fds; ++fl)
+				{
+					if ( fileno(stdout) == fl && use_output )
+						if ( strcmp(output, "-") == 0 )
+							continue;
+					close(fl);
+				}
+ 
+				setsid();
+			}
+			else
+				exit(0);
 		}
 		else
-			exit(0);
+		{
+			fputs("klogd: Already running.\n", stderr);
+			exit(1);
+		}
 	}
-#endif
 
 
+	/* tuck my process id away */
+	if (!check_pid(PidFile))
+	{
+		if (!write_pid(PidFile))
+			Terminate();
+	}
+	else
+	{
+		fputs("klogd: Already running.\n", stderr);
+		Terminate();
+	}
+	
+
 	/* Signal setups. */
 	for (ch= 1; ch < NSIG; ++ch)
 		signal(ch, SIG_IGN);
@@ -483,9 +681,11 @@ int main(argc, argv)
 
 
 	/* Open outputs. */
-	if ( output != (char *) 0 )
+	if ( use_output )
 	{
-		if ( (output_file = fopen(output, "w")) == (FILE *) 0 )
+		if ( strcmp(output, "-") == 0 )
+			output_file = stdout;
+		else if ( (output_file = fopen(output, "w")) == (FILE *) 0 )
 		{
 			fprintf(stderr, "klogd: Cannot open output file %s - "\
 				"%s\n", output, strerror(errno));
@@ -499,26 +699,26 @@ int main(argc, argv)
 	/* Handle one-shot logging. */
 	if ( one_shot )
 	{
+		InitKsyms(symfile);
 		if ( (logsrc = GetKernelLogSrc()) == kernel )
 			LogKernelLine();
 		else
 			LogProcLine();
-		stop_daemon(0);
-	}
-
-	/* tuck my process id away */
-	fp = fopen(PidFile, "w");
-	if (fp != NULL) {
-		fprintf(fp, "%d\n", getpid());
-		(void) fclose(fp);
+		Terminate();
 	}
 
 	/* Determine where kernel logging information is to come from. */
-	sleep(1);
+#if defined(KLOGD_DELAY)
+	sleep(KLOGD_DELAY);
+#endif
 	logsrc = GetKernelLogSrc();
+	InitKsyms(symfile);
 
         /* The main loop. */
 	while (1)
+	{
+		if ( change_state )
+			ChangeLogging();
 		switch ( logsrc )
 		{
 			case kernel:
@@ -528,6 +728,8 @@ int main(argc, argv)
 				LogProcLine();
 				break;
 		        case none:
+				pause();
 				break;
 		}
+	}
 }
