@@ -39,6 +39,11 @@ static char sccsid[] = "@(#)syslog.c	5.28 (Berkeley) 6/27/90";
  * Sat Dec 11 11:58:31 CST 1993: Dr. Wettstein
  *	Changes to allow compilation with no complains under -Wall.
  *
+ * Thu Jan 18 11:16:11 CST 1996: Dr. Wettstein
+ *	Added patch to close potential security hole.  This is the same
+ *	patch which was announced in the linux-security mailing lists
+ *	and incorporated into the libc version of syslog.c.
+ *
  */
 
 #include <sys/types.h>
@@ -51,11 +56,7 @@ static char sccsid[] = "@(#)syslog.c	5.28 (Berkeley) 6/27/90";
 #include "pathnames.h"
 #endif
 
-#ifdef	LINUX
-#include <linux/uio.h>
-#else
 #include <sys/uio.h>
-#endif
 #include <sys/wait.h>
 #include <netdb.h>
 #include <string.h>
@@ -132,7 +133,9 @@ vsyslog(pri, fmt, ap)
 		register char ch, *t1, *t2;
 		char *strerror();
 
-		for (t1 = fmt_cpy; (ch = *fmt); ++fmt)
+		for (t1 = fmt_cpy;
+		     (ch = *fmt) != '\0' && t1<fmt_cpy+sizeof(fmt_cpy);
+		     ++fmt)
 			if (ch == '%' && fmt[1] == 'm') {
 				++fmt;
 				for (t2 = strerror(saved_errno);
